/**
 * SrbnFreaks(SOROBAN FREAKS)
 * (c) Kenta Cho (cs8k-cyu@asahi-net.or.jp)
 * http://www.asahi-net.or.jp/~cs8k-cyu/
 *
 * soroban.c(main routine)
 */

#include <Common.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include "soroban.h"

// bitmap handlers
static VoidHand objectBitmapHandles[bitmapTypeCount];
static BitmapPtr objectBitmapPtr[bitmapTypeCount];
static WinHandle objectWindowHandles[bitmapTypeCount];

static GameStatus gameStatus;
static int total;

#include "drawsoroban.c"
#include "gamemanager.c"

/**
 * load bitmaps and hold these handlers
 */
static void loadBitmaps() {
  int i;
  WinHandle oldDrawWinH;
  Word error;
  oldDrawWinH = WinGetDrawWindow();
  for ( i=0; i<bitmapTypeCount; i++ ) {
    objectBitmapHandles[i] = DmGetResource(bitmapRsc, FIRST_BITMAP_ID+i);
    objectBitmapPtr[i] = MemHandleLock(objectBitmapHandles[i]);
    objectWindowHandles[i] = WinCreateOffscreenWindow
      (objectBitmapPtr[i]->width, objectBitmapPtr[i]->height,
       screenFormat, &error);
    ErrFatalDisplayIf(error, "Error loading images");
    WinSetDrawWindow(objectWindowHandles[i]);
    WinDrawBitmap(objectBitmapPtr[i], 0, 0);
  }
  WinSetDrawWindow(oldDrawWinH);
}

static int StartApplication(void) {
  randomize();
  loadBitmaps();
  initGame();
}

/**
 * relase bitmap handlers
 */
static void unloadBitmaps() {
  int i;
  for ( i=0; i<bitmapTypeCount; i++ ) {
    MemPtrUnlock(objectBitmapPtr[i]);
    DmReleaseResource(objectBitmapHandles[i]);
    if (objectWindowHandles[i]) 
      WinDeleteWindow(objectWindowHandles[i], false);
  }
}

static void StopApplication(void) {
  unloadBitmaps();
}

/**
 * event hanlder for title form
 */
static Boolean TitleFormHandleEvent(EventPtr event) {
  Boolean handled = FALSE;
  switch ( event->eType ) {
  case ctlSelectEvent:
    if ( event->data.ctlEnter.controlID == PRO_START_BUTTON ) {
      startButtonPushed(PRO_MODE);
      handled = TRUE;
    } else if ( event->data.ctlEnter.controlID == AMA_START_BUTTON ) {
      startButtonPushed(AMA_MODE);
      handled = TRUE;
    }
    break;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    if ( score >= 0 ) drawResult();
    handled = TRUE;
    break;
  }
  return handled;
}

static int lineIdx, topDamaIdx, btmDamaIdx;
static int sx, sy;
static int fpy;

/**
 * check pen down position
 */
static void setFirstPenPos(EventPtr event) {
  int i;
  lineIdx = -1;
  if ( gameStatus.questionCnd == START ) return;
  sx = event->screenX;
  sy = fpy = event->screenY; 
  for ( i=0 ; i<SOROBAN_LINE_NMB ; i++ ) {
    if ( sx>=sl[i].x-4 && sx<sl[i].x+DAMA_WIDTH+4 ) {
      lineIdx = i; 
      topDamaIdx = (fpy-sl[i].topy)/DAMA_HALF_HEIGHT;
      if ( topDamaIdx > 6 ) topDamaIdx = -999;
      else if ( topDamaIdx < -1 ) topDamaIdx = -1;
      else if ( topDamaIdx > 4 ) topDamaIdx = 4;
      btmDamaIdx = (fpy-sl[i].btmy)/DAMA_HALF_HEIGHT;
      if ( btmDamaIdx < -2 ) btmDamaIdx = -999;
      else if ( btmDamaIdx < -1 ) btmDamaIdx = -1;
      else if ( btmDamaIdx > 10 ) btmDamaIdx = 10;
      break;
    }
  }
}

static void resetPenPos() {
  lineIdx = -1;
}

/**
 * check pen position
 */
static void penMoved(int px, int py) {
  sx = px; sy = py;
  if ( gameStatus.waitTick > 0 ) {
    return;
  } else if ( gameStatus.questionCnd == START ) {
    int i, uc = 0;
    for ( i=0 ; i<SOROBAN_LINE_NMB ; i++ ) {
      if ( sl[i].topIdx == 0 ) {
	if ( sx >= sl[i].x && sx < sl[i].x+DAMA_WIDTH
	    && sy > sl[i].topy+DAMA_HEIGHT 
	    && sy < sl[i].topy+DAMA_HEIGHT*2 ) {
	  moveTopDama(i, 1, 0);
	  drawStartIndicator();
	  uc++;
	}
      } else {
	uc++;
      }
    }
    if ( uc >= SOROBAN_LINE_NMB ) {
      gameStatus.questionCnd = GOTO_QUESTION;
      score++;
    }
  } else {
    int ofy;
    int idx = lineIdx;
    if ( lineIdx == -1 ) return;
    ofy = sy-fpy;
    if ( ofy > DAMA_HEIGHT/8 ) {
      if ( btmDamaIdx != -999 ) {
	if ( moveBtmDama(idx, (btmDamaIdx+1)/2, 1) ) {
	  resetPenPos();
	  return;
	}
      }
      if ( topDamaIdx != -999 ) {
	if ( moveTopDama(idx, (topDamaIdx+1)/2, 1) ) {
	  resetPenPos();
	  return;
	}
      }
    } else if ( ofy < -DAMA_HEIGHT/8 ) {
      if ( topDamaIdx != -999 ) {
	if ( moveTopDama(idx, (topDamaIdx-1)/2, 0) ) {
	  resetPenPos();
	  return;
	}
      }
      if ( btmDamaIdx != -999 ) {
	if ( moveBtmDama(idx, (btmDamaIdx-1)/2, 0) ) {
	  resetPenPos();
	  return;
	}
      }
    }
  }
}

static void handlePenMove(EventPtr event) {
  penMoved(event->screenX, event->screenY);
}

/**
 * event hanlder for main form(soroban form)
 */
static Boolean MyFormHandleEvent(EventPtr event) {
  Boolean handled = FALSE;
  long nt;
  SWord px, py;
  Boolean pd;
  switch ( event->eType ) {
  case penDownEvent:
    setFirstPenPos(event);
    handled = TRUE;
    break;
  case penMoveEvent:
  case penUpEvent:
    handlePenMove(event);
    handled = TRUE;
    break;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    startGame();
    handled = TRUE;
    break;
  default:
    EvtGetPen(&px, &py, &pd);
    if ( pd && (sx!=px || sy!=py) ) {
      penMoved(px, py);
    }
    handled = TRUE;
    break;
  }
  return handled;
}

/**
 * activate forms
 */
static Boolean ApplicationHandleEvent(EventPtr event) {
  FormPtr frm;
  Int formID;
  Boolean handled = FALSE;

  if ( event->eType == frmLoadEvent ) {
    formID = event->data.frmLoad.formID;
    frm = FrmInitForm(formID);
    FrmSetActiveForm(frm);
    switch ( formID ) {
    case SOROBAN_FORM:
      FrmSetEventHandler(frm, MyFormHandleEvent);
      break;
    case TITLE_FORM:
      FrmSetEventHandler(frm, TitleFormHandleEvent);
      break;
    }
    handled = TRUE;
  }
  return handled;
}

// event loop interval
#define INTERVAL 20

static void EventLoop(void) {
  EventType event;
  Word error;
  do {
    EvtGetEvent(&event, INTERVAL);
    if ( !SysHandleEvent(&event) ) {
      if ( !MenuHandleEvent(0, &event, &error) ) {
	if ( !ApplicationHandleEvent(&event) ) {
	  FrmDispatchEvent(&event);
	}
      }
    }
    handleGameStatus();
  } while ( event.eType != appStopEvent );
}

DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags) {
  if ( cmd != sysAppLaunchCmdNormalLaunch ) return 0;
  StartApplication();
  EventLoop();
  StopApplication();
  return 0;
}
